package at.ac.uibk.gitsearch.security.oauth2;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.springframework.security.oauth2.client.web.AuthorizationRequestRepository;
import org.springframework.security.oauth2.core.endpoint.OAuth2AuthorizationRequest;
import org.springframework.security.oauth2.core.endpoint.OAuth2ParameterNames;
import org.springframework.util.Assert;

/**
 * a specialized version of OAuth2 2AuthorizationRequestRepository that does not uses sessions to store requests.
 * @author Michael Breu
 *
 */
public class GitSearchOAuth2AuthorizationRequestRepository implements AuthorizationRequestRepository<OAuth2AuthorizationRequest> {

    private static final String REQUEST_CANNOT_BE_NULL = "request cannot be null";
    public static final String REFERER_ATTRIBUTE = "referer";
    /**
     * TODO this should be persisted in the database!
     */
    private final Map<String, OAuth2AuthorizationRequest> authorizationRequests = new ConcurrentHashMap<>();

    @Override
    public OAuth2AuthorizationRequest loadAuthorizationRequest(HttpServletRequest request) {
        Assert.notNull(request, REQUEST_CANNOT_BE_NULL);
        String stateParameter = this.getStateParameter(request);
        if (stateParameter == null) {
            return null;
        }
        return authorizationRequests.get(stateParameter);
    }

    @Override
    public void saveAuthorizationRequest(
        OAuth2AuthorizationRequest authorizationRequest,
        HttpServletRequest request,
        HttpServletResponse response
    ) {
        Assert.notNull(request, REQUEST_CANNOT_BE_NULL);
        Assert.notNull(response, "response cannot be null");
        if (authorizationRequest == null) {
            this.removeAuthorizationRequest(request, response);
            return;
        }
        String state = authorizationRequest.getState();
        Assert.hasText(state, "authorizationRequest.state cannot be empty");
        String referer = request.getHeader(REFERER_ATTRIBUTE);
        if (referer == null) {
            referer = "/";
        }
        final Map<String, Object> requestAttributes = new ConcurrentHashMap<>(authorizationRequest.getAttributes());
        requestAttributes.put(REFERER_ATTRIBUTE, referer);
        OAuth2AuthorizationRequest extendedAuthorizationRequest = OAuth2AuthorizationRequest
            .from(authorizationRequest)
            .attributes(requestAttributes)
            .build();
        authorizationRequests.put(state, extendedAuthorizationRequest);
    }

    @Override
    public OAuth2AuthorizationRequest removeAuthorizationRequest(HttpServletRequest request) {
        Assert.notNull(request, REQUEST_CANNOT_BE_NULL);
        String stateParameter = this.getStateParameter(request);
        if (stateParameter == null) {
            return null;
        }
        return authorizationRequests.remove(stateParameter);
    }

    @Override
    public OAuth2AuthorizationRequest removeAuthorizationRequest(HttpServletRequest request, HttpServletResponse response) {
        Assert.notNull(response, "response cannot be null");
        return this.removeAuthorizationRequest(request);
    }

    /**
     * Gets the state parameter from the {@link HttpServletRequest}
     * @param request the request to use
     * @return the state parameter or null if not found
     */
    private String getStateParameter(HttpServletRequest request) {
        return request.getParameter(OAuth2ParameterNames.STATE);
    }
}
