package at.ac.uibk.gitsearch.web.rest;

import static org.assertj.core.api.Assertions.assertThat;
import static org.hamcrest.Matchers.hasSize;
import static org.junit.Assert.assertThrows;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.springframework.security.test.web.servlet.request.SecurityMockMvcRequestPostProcessors.csrf;
import static org.springframework.test.web.servlet.result.MockMvcResultHandlers.print;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.jsonPath;

import at.ac.uibk.gitsearch.GitsearchApp;
import at.ac.uibk.gitsearch.domain.Review;
import at.ac.uibk.gitsearch.domain.ReviewComment;
import at.ac.uibk.gitsearch.domain.ReviewRating;
import at.ac.uibk.gitsearch.domain.Statistics;
import at.ac.uibk.gitsearch.domain.User;
import at.ac.uibk.gitsearch.domain.enumeration.ReviewStatus;
import at.ac.uibk.gitsearch.repository.jpa.ReviewJPARepository;
import at.ac.uibk.gitsearch.repository.jpa.ReviewRatingJPARepository;
import at.ac.uibk.gitsearch.repository.jpa.StatisticsRepository;
import at.ac.uibk.gitsearch.security.AuthoritiesConstants;
import at.ac.uibk.gitsearch.service.ReviewService;
import at.ac.uibk.gitsearch.service.UserService;
import at.ac.uibk.gitsearch.service.dto.AdminUserDTO;
import at.ac.uibk.gitsearch.service.dto.ReviewDTO;
import at.ac.uibk.gitsearch.service.dto.ReviewRatingDTO.ReviewStatisticsDTO;
import at.ac.uibk.gitsearch.service.dto.ReviewRequest;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.TimeUnit;
import javax.persistence.EntityManager;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.transaction.Transactional;
import org.hamcrest.Matchers;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.http.MediaType;
import org.springframework.mock.web.MockHttpServletRequest;
import org.springframework.mock.web.MockHttpServletResponse;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.core.userdetails.UsernameNotFoundException;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.security.test.web.support.WebTestUtils;
import org.springframework.security.web.context.HttpRequestResponseHolder;
import org.springframework.security.web.context.SecurityContextRepository;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.request.MockHttpServletRequestBuilder;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;
import org.springframework.test.web.servlet.request.RequestPostProcessor;

@SpringBootTest(classes = GitsearchApp.class)
@ExtendWith(MockitoExtension.class)
@AutoConfigureMockMvc
@WithMockUser
class ReviewResourceIT {

    private static final String DEFAULT_LINK = "AAAAAAAAAA";
    private static final String UPDATED_LINK = "BBBBBBBBBB";

    private static final Long DEFAULT_REQUESTED_BY = 1L;
    private static final Long UPDATED_REQUESTED_BY = 2L;

    @SuppressWarnings("unused")
    private static final Long SMALLER_REQUESTED_BY = 1L - 1L;

    private static final String DEFAULT_RESOURCE = "AAAAAAAAAA";
    private static final String UPDATED_RESOURCE = "BBBBBBBBBB";

    private static final String DEFAULT_RESOURCE_ID = "AAAAAAAAAA";
    private static final String UPDATED_RESOURCE_ID = "BBBBBBBBBB";

    private static final String ENTITY_API_URL = "/api/reviews";

    @SuppressWarnings("unused")
    private static final String ENTITY_API_URL_ID = ENTITY_API_URL + "/{id}";

    @Autowired
    private MockMvc mockMvc;

    @Autowired
    private ReviewJPARepository reviewRepository;

    @Autowired
    private ReviewRatingJPARepository reviewRatingRepository;

    @Autowired
    private ReviewService reviewService;

    @Autowired
    private UserService userService;

    @Autowired
    private StatisticsRepository statisticsRepository;

    @BeforeEach
    public void tearDown() {
        reviewRatingRepository.deleteAll();
        reviewRepository.deleteAll();
    }

    public static int resourceCounter = 0; // just to make test resources unique :-(

    /**
     * Create an entity for this test.
     *
     * This is a static method, as tests for other entities might also need it,
     * if they test an entity which requires the current entity.
     */
    public static Review createEntity(EntityManager em) {
        Review review = new Review()
            .link(DEFAULT_LINK)
            .requestedBy(DEFAULT_REQUESTED_BY)
            .resource(DEFAULT_RESOURCE + (resourceCounter++))
            .resourceID(DEFAULT_RESOURCE_ID);
        return review;
    }

    /**
     * Create an updated entity for this test.
     *
     * This is a static method, as tests for other entities might also need it,
     * if they test an entity which requires the current entity.
     */
    public static Review createUpdatedEntity(EntityManager em) {
        Review review = new Review()
            .link(UPDATED_LINK)
            .requestedBy(UPDATED_REQUESTED_BY)
            .resource(UPDATED_RESOURCE)
            .resourceID(UPDATED_RESOURCE_ID);
        return review;
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN, username = "admin")
    @Transactional
    void createReviewTest() throws Exception {
        ReviewDTO reviewDTO = new ReviewDTO();
        reviewDTO.setId(1L);
        reviewDTO.setResource("resource");
        User user = userService.getUserWithAuthorities().get();

        List<String> reviewers = new ArrayList<>();
        reviewers.add("admin@localhost");

        reviewDTO.setUsers(reviewers);

        reviewDTO.setResource("resource");

        reviewDTO.setBadgeRewarded(false);

        List<String> comm = new ArrayList<>();
        comm.add("comment");

        ReviewComment sampleComment = new ReviewComment();
        sampleComment.setDescriptionIsClear(true);
        sampleComment.setRequirementsAreClear(false);
        sampleComment.setDescriptionComment("The description is clear, but the requirements are not.");

        sampleComment.setInformationRequiredComplete(true);
        sampleComment.setAllMaterialsAvailable(false);
        sampleComment.setInformationRequiredComment("All required information is provided, but not all materials are available.");

        sampleComment.setIsStructured(true);
        sampleComment.setSubExercisesAreClear(false);
        sampleComment.setStructuredComment("The structure is good, but the sub-exercises are not clear.");

        sampleComment.setSolutionIsCorrect(true);
        sampleComment.setSolutionComment("The solution is correct.");

        sampleComment.setMetadataIsComplete(false);
        sampleComment.setMetadataIsCorrect(true);
        sampleComment.setMetadataComment("The metadata is correct but not complete.");

        reviewDTO.setComments(Collections.singletonList(sampleComment));

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .post("/api/review/create")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewDTO))
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        // Validate the Review is in the database
        List<Review> reviewList = reviewRepository.findAll();
        assertThat(reviewList).hasSize(1);
        Review testReview = reviewList.get(0);
        assertThat(testReview.getResource()).isEqualTo("resource");
        assertThat(testReview.getRequestedBy()).isEqualTo(user.getId());

        ReviewRating reviewRating = reviewRatingRepository.findAll().get(0);
        assertThat(reviewRating.getUser().getEmail()).isEqualTo("admin@localhost");
        assertThat(reviewRating.getReview().getResource()).isEqualTo("resource");
        assertThat(reviewRating.getStatus()).isEqualTo(ReviewStatus.REVIEW_IN_PROGRESS);
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN, username = "admin")
    @Transactional
    void updateReviewTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("user@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));

        // Validate the Revoew is in the database
        List<Review> reviewList = reviewRepository.findAll();
        assertThat(reviewList).hasSize(1);

        Review testReview = reviewList.get(0);
        assertThat(testReview.getResource()).isEqualTo("resource");
        testReview.setResource("resource2");

        ReviewDTO reviewDTO = reviewService.reviewToDTO(testReview);

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .put("/api/review")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewDTO))
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        assertThat(reviewService.getReview(testReview.getId()).getResource()).isEqualTo("resource2");
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    @Transactional
    void updateReviewRatingByUserTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("user@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));
        // Validate the Revoew is in the database
        List<ReviewRating> reviewList = reviewRatingRepository.findAllByUserId(userService.getUserWithAuthorities().get().getId());
        assertThat(reviewList).hasSize(1);

        ReviewRating testReview = reviewList.get(0);
        assertThat(testReview.getReview().getResource()).isEqualTo("resource");
        assertThat(testReview.getStatus()).isEqualTo(ReviewStatus.REVIEW_IN_PROGRESS);

        ReviewDTO reviewDTO = reviewService.reviewToDTO(testReview.getReview());
        List<String> status = new ArrayList<>();
        status.add("DELETION_REQUESTED");

        reviewDTO.setStatus(status);

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .put("/api/review/updateByUser")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewDTO))
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        List<ReviewDTO> reviews = reviewService.getAllReviews();

        assertEquals(0, reviews.size());
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN, username = "admin")
    @Transactional
    void getReviewTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("user@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));

        List<Review> reviewList = reviewRepository.findAll();
        Review testReview = reviewList.get(0);

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .get("/api/review/{id}", testReview.getId())
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        String str = result.getResponse().getContentAsString();

        ReviewDTO reviewDTO = new ObjectMapper().readValue(str, ReviewDTO.class);

        assertThat(reviewDTO.getResource()).isEqualTo("resource");
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN, username = "admin")
    @Transactional
    void getAllReviewsTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("user@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));
        reviewService.createReview(new ReviewRequest("resource2", "codeability2.at", "2", reviewers));

        MvcResult result = mockMvc
            .perform(MockMvcRequestBuilders.get("/api/review/all").with(csrf().asHeader()).contentType(MediaType.APPLICATION_JSON))
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        String str = result.getResponse().getContentAsString();

        List<ReviewDTO> reviewDTOs = new ObjectMapper().readValue(str, new TypeReference<List<ReviewDTO>>() {});

        assertThat(reviewDTOs).hasSize(2);
        assertThat(reviewDTOs.get(0).getResource()).isEqualTo("resource");
        assertThat(reviewDTOs.get(1).getResource()).isEqualTo("resource2");
        assertThat(reviewDTOs.get(0).getRequestedBy()).isEqualTo("admin@localhost");
        assertThat(reviewDTOs.get(1).getRequestedBy()).isEqualTo("admin@localhost");
        assertThat(reviewDTOs.get(0).getStatus().get(0)).isEqualTo(ReviewStatus.REVIEW_IN_PROGRESS.name());
        assertThat(reviewDTOs.get(1).getStatus().get(0)).isEqualTo(ReviewStatus.REVIEW_IN_PROGRESS.name());
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN, username = "admin")
    @Transactional
    void deleteReviewTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("user@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));

        List<Review> reviewList = reviewRepository.findAll();

        Review testReview = reviewList.get(0);

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .delete("/api/review/{id}", testReview.getId())
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        assertThat(reviewRepository.findAll()).isEmpty();
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    @Transactional
    void getReviewsByUser() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("user@localhost");
        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));

        MvcResult result = mockMvc
            .perform(MockMvcRequestBuilders.get("/api/review/allByUser").with(csrf().asHeader()).contentType(MediaType.APPLICATION_JSON))
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        String str = result.getResponse().getContentAsString();

        List<ReviewDTO> reviewDTOs = new ObjectMapper().readValue(str, new TypeReference<List<ReviewDTO>>() {});

        assertThat(reviewDTOs.get(0).getResource()).isEqualTo("resource");
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    @Transactional
    void requestReviewTest() throws Exception {
        ReviewRequest reviewRequest = new ReviewRequest("resource", "codeability.at", "1", new ArrayList<>());

        User user = userService.getUserWithAuthorities().get();

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .post("/api/requestReview")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewRequest))
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        List<Review> reviewList = reviewRepository.findAll();

        Review testReview = reviewList.get(0);

        assertThat(testReview.getResource()).isEqualTo("resource");
        assertThat(testReview.getRequestedBy()).isEqualTo(user.getId());
        assertThat(testReview.getComments()).isEmpty();
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    @Transactional
    void notifyReviewersTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("admin@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));

        User user = userService.getUserWithAuthorities().get();

        List<Review> reviewList = reviewRepository.findAll();

        Review testReview = reviewList.get(0);

        ReviewDTO reviewDTO = reviewService.reviewToDTO(testReview);

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .post("/api/review/notifyReviewers")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewDTO))
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        reviewList = reviewRepository.findAll();

        testReview = reviewList.get(0);

        assertThat(testReview.getResource()).isEqualTo("resource");
        assertThat(testReview.getRequestedBy()).isEqualTo(user.getId());
        assertThat(testReview.getComments().size()).isOne();

        ReviewRating reviewRating = testReview.getComments().stream().findFirst().get();

        assertThat(reviewRating.getStatus()).isEqualTo(ReviewStatus.REVIEW_IMPROVED);
        assertThat(reviewRating.getReview().getId()).isEqualTo(testReview.getId());
        assertThat(reviewRating.getUser().getEmail()).isEqualTo("admin@localhost");
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN, username = "admin")
    void rewardBadgeTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("admin@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));

        List<Review> reviewList = reviewRepository.findAll();

        Review testReview = reviewList.get(0);

        ReviewDTO reviewDTO = reviewService.reviewToDTO(testReview);

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .post("/api/review/rewardBadge")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewDTO))
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        reviewList = reviewRepository.findAll();

        testReview = reviewList.get(0);

        assertThat(testReview.getResource()).isEqualTo("resource");

        ReviewRating reviewRating = testReview.getComments().stream().findFirst().get();

        assertThat(reviewRating.getStatus()).isEqualTo(ReviewStatus.REVIEW_COMPLETED);
        assertThat(reviewRating.getReview().getId()).isEqualTo(testReview.getId());
        assertThat(reviewRating.getUser().getEmail()).isEqualTo("admin@localhost");

        Optional<Statistics> statistics = statisticsRepository.findByExerciseID(testReview.getResourceID());

        assertThat(statistics).isPresent();
        assertThat(statistics.get().getBadgeRewarded()).isTrue();
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    @Transactional
    void changeStatusTest() throws Exception {
        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .post("/api/review/changeStatus")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(true))
            )
            .andReturn();

        User user = userService.getUserByLogin("user").get();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);
        assertThat(user.isReviewingEnabled()).isTrue();

        result =
            mockMvc
                .perform(
                    MockMvcRequestBuilders
                        .post("/api/review/changeStatus")
                        .with(csrf().asHeader())
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(TestUtil.convertObjectToJsonBytes(false))
                )
                .andReturn();

        user = userService.getUserByLogin("user").get();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);
        assertThat(user.isReviewingEnabled()).isFalse();
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN, username = "admin")
    @Transactional
    void allWithStatisticsTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("admin@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));
        reviewService.createReview(new ReviewRequest("resource1", "codeability1.at", "2", reviewers));
        reviewService.createReview(new ReviewRequest("resource2", "codeability2.at", "3", reviewers));

        List<Review> reviewList = reviewRepository.findAll();

        reviewList.forEach(review -> {
            reviewService.rewardBadge(review.getId());
        });

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders.get("/api/review/allWithStatistics").with(csrf().asHeader()).contentType(MediaType.APPLICATION_JSON)
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        String str = result.getResponse().getContentAsString();

        List<ReviewDTO> reviewDTOs = new ObjectMapper().readValue(str, new TypeReference<List<ReviewDTO>>() {});

        reviewDTOs.forEach(reviewDTO -> {
            assertThat(reviewDTO.getBadgeRewarded()).isTrue();
        });
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    @Transactional
    void countAllByUserTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("user@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));
        reviewService.createReview(new ReviewRequest("resource1", "codeability1.at", "2", reviewers));
        reviewService.createReview(new ReviewRequest("resource2", "codeability2.at", "3", reviewers));

        List<Review> reviewList = reviewRepository.findAll();

        reviewList.forEach(review -> {
            reviewService.rewardBadge(review.getId());
        });

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders.post("/api/review/countAllByUser").with(csrf().asHeader()).contentType(MediaType.APPLICATION_JSON)
            )
            .andDo(print())
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        String str = result.getResponse().getContentAsString();

        ReviewStatisticsDTO stats = new ObjectMapper().readValue(str, ReviewStatisticsDTO.class);
        assertThat(stats.getReviewCompleted())
            .withFailMessage("Status not as expected. This may result from changed testdata")
            .isEqualTo(3);
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    @Transactional
    void checkUnknownEmail() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("reviewer@localhost");
        final ReviewRequest reviewRequest = new ReviewRequest("resource", "codeability.at", "1", reviewers);

        assertThrows(
            UsernameNotFoundException.class,
            () -> {
                reviewService.createReview(reviewRequest);
            }
        );
    }

    @Test
    @Timeout(value = 10, unit = TimeUnit.MINUTES)
    @Transactional
    void getReviewHistory() throws Exception {
        VariableTestSecurityContextHolderPostProcessor testUserContext = new VariableTestSecurityContextHolderPostProcessor(
            new String[] { "ROLE_USER" },
            "testuser"
        );
        VariableTestSecurityContextHolderPostProcessor reviewerContext = new VariableTestSecurityContextHolderPostProcessor(
            new String[] { "ROLE_USER" },
            "testreviewer"
        );
        VariableTestSecurityContextHolderPostProcessor adminContext = new VariableTestSecurityContextHolderPostProcessor(
            new String[] { "ROLE_ADMIN", "ROLE_USER" },
            "testadmin"
        );
        ReviewRequest reviewRequest = new ReviewRequest("testResource", "codeability.at", "1", new ArrayList<>());

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .post("/api/requestReview")
                    .with(csrf().asHeader())
                    .with(testUserContext)
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewRequest))
            )
            .andReturn();
        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        testUserContext.setSecurityContext();
        final List<ReviewDTO> userReviews = reviewService.getReviewsByUser();

        assertThat(userReviews).hasSize(1);

        ReviewDTO userReview = userReviews.get(0);

        MvcResult resultHistory = getHistoryFor(testUserContext, userReview)
            .andExpect(jsonPath("$[0].review").value(userReview.getId()))
            .andExpect(jsonPath("$[0].user").value(0)) // anonymized
            .andReturn();

        assertThat(resultHistory.getResponse().getStatus()).isEqualTo(200);

        MvcResult resultHistory2 = getHistoryFor(reviewerContext, userReview).andExpect(jsonPath("$", Matchers.empty())).andReturn();

        assertThat(resultHistory2.getResponse().getStatus()).isEqualTo(200);

        // ------------------ now add some reviewers

        userReview.setUsers(Collections.singletonList(reviewerContext.getMockedUser().getEmail()));

        MvcResult resultUpdate = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .put("/api/review")
                    .with(csrf().asHeader())
                    .with(adminContext)
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(userReview))
            )
            .andReturn();
        assertThat(resultUpdate.getResponse().getStatus()).isEqualTo(200);

        MvcResult resultHistoryForUser = getHistoryFor(testUserContext, userReview)
            .andExpect(jsonPath("$").isArray())
            .andExpect(jsonPath("$", hasSize(2)))
            .andExpect(jsonPath("$[0].review").value(userReview.getId()))
            .andExpect(jsonPath("$[0].user").value(0)) // anonymized
            .andReturn();

        assertThat(resultHistoryForUser.getResponse().getStatus()).isEqualTo(200);

        MvcResult resultHistoryForReviewer = getHistoryFor(reviewerContext, userReview)
            .andExpect(jsonPath("$").isArray())
            .andExpect(jsonPath("$", hasSize(1)))
            .andExpect(jsonPath("$[0].review").value(userReview.getId()))
            .andExpect(jsonPath("$[0].user").value(0)) // anonymized
            .andReturn();

        assertThat(resultHistoryForReviewer.getResponse().getStatus()).isEqualTo(200);

        // Admin user will see nothing, because he does not act as a reviewer or
        // requester
        MvcResult resultHistoryAdminUser = getHistoryFor(adminContext, userReview).andExpect(jsonPath("$", Matchers.empty())).andReturn();

        assertThat(resultHistoryAdminUser.getResponse().getStatus()).isEqualTo(200);

        // additionally with reviewRatingId
        final List<ReviewRating> reviewRatings = reviewRatingRepository.findByReview(userReview.getId());
        assertThat(reviewRatings).hasSize(1);
        ReviewRating reviewRating = reviewRatings.get(0);

        resultHistoryForUser =
            getHistoryFor(testUserContext, userReview, reviewRating.getId())
                .andExpect(jsonPath("$").isArray())
                .andExpect(jsonPath("$", hasSize(1)))
                .andExpect(jsonPath("$[0].review").value(userReview.getId()))
                .andExpect(jsonPath("$[0].user").value(0)) // anonymized
                .andReturn();

        assertThat(resultHistoryForUser.getResponse().getStatus()).isEqualTo(200);

        resultHistoryForReviewer =
            getHistoryFor(reviewerContext, userReview, reviewRating.getId())
                .andExpect(jsonPath("$").isArray())
                .andExpect(jsonPath("$", hasSize(1)))
                .andExpect(jsonPath("$[0].review").value(userReview.getId()))
                .andExpect(jsonPath("$[0].user").value(0)) // anonymized
                .andReturn();

        assertThat(resultHistoryForReviewer.getResponse().getStatus()).isEqualTo(200);

        // Admin user will see nothing, because he does not act as a reviewer or
        // requester
        resultHistoryAdminUser =
            getHistoryFor(adminContext, userReview, reviewRating.getId()).andExpect(jsonPath("$", Matchers.empty())).andReturn();

        assertThat(resultHistoryAdminUser.getResponse().getStatus()).isEqualTo(200);
    }

    private ResultActions getHistoryFor(VariableTestSecurityContextHolderPostProcessor testUserContext, ReviewDTO userReview)
        throws Exception {
        return getHistoryFor(testUserContext, userReview, null);
    }

    private ResultActions getHistoryFor(
        VariableTestSecurityContextHolderPostProcessor testUserContext,
        ReviewDTO userReview,
        Long reviewRatingId
    ) throws Exception {
        MockHttpServletRequestBuilder requestBuilder = MockMvcRequestBuilders
            .get("/api/review/reviewHistory")
            .param("reviewId", userReview.getId().toString())
            .with(csrf().asHeader())
            .with(testUserContext);
        if (reviewRatingId != null) {
            requestBuilder = requestBuilder.param("reviewRatingId", reviewRatingId.toString());
        }
        return mockMvc.perform(requestBuilder).andDo(print());
    }

    public class VariableTestSecurityContextHolderPostProcessor implements RequestPostProcessor {

        SecurityContext mockedContext;
        User mockedUser;

        protected VariableTestSecurityContextHolderPostProcessor(String[] authorities, String username) {
            super();
            mockedContext = withMockUser(authorities, username);
            mockedUser = createTestUser(username);
        }

        private User createTestUser(String userName) {
            AdminUserDTO newUser = new AdminUserDTO();
            newUser.setEmail(userName + "@codability.org");
            newUser.setFirstName("First " + userName);
            newUser.setLastName("Last " + userName);
            newUser.setLogin(userName);
            User createdUser = userService.createUser(newUser);
            return createdUser;
        }

        @Override
        public MockHttpServletRequest postProcessRequest(MockHttpServletRequest request) {
            //            SecurityContext existingContext = TestSecurityContextRepository.getContext(request);
            SecurityContext existingContext = mockedContext;
            save(existingContext, request);
            return request;
        }

        public User getMockedUser() {
            return mockedUser;
        }

        /**
         * Saves the {@link SecurityContext} using the {@link SecurityContextRepository}
         *
         * @param securityContext the {@link SecurityContext} to save
         * @param request         the {@link HttpServletRequest} to use
         */
        void save(SecurityContext securityContext, HttpServletRequest request) {
            SecurityContextRepository securityContextRepository = WebTestUtils.getSecurityContextRepository(request);
            WebTestUtils.setSecurityContextRepository(request, securityContextRepository);
            HttpServletResponse response = new MockHttpServletResponse();
            HttpRequestResponseHolder requestResponseHolder = new HttpRequestResponseHolder(request, response);
            securityContextRepository.loadContext(requestResponseHolder);
            request = requestResponseHolder.getRequest();
            response = requestResponseHolder.getResponse();
            securityContextRepository.saveContext(securityContext, request, response);
        }

        void setSecurityContext() {
            SecurityContextHolder.setContext(mockedContext);
        }

        private SecurityContext withMockUser(String[] authorities, String username) {
            List<GrantedAuthority> grantedAuthorities = new ArrayList<>();
            for (String authority : authorities) {
                grantedAuthorities.add(new SimpleGrantedAuthority(authority));
            }
            org.springframework.security.core.userdetails.User principal = new org.springframework.security.core.userdetails.User(
                username,
                "somePassword",
                true,
                true,
                true,
                true,
                grantedAuthorities
            );
            Authentication authentication = new UsernamePasswordAuthenticationToken(
                principal,
                principal.getPassword(),
                principal.getAuthorities()
            );
            SecurityContext context = SecurityContextHolder.createEmptyContext();
            context.setAuthentication(authentication);
            //            SecurityContextHolder.setContext(context);
            return context;
        }
    }
}
